#!D:\bld\pyarrow_1572781486807\_h_env\python.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'pyarrow==0.15.1','console_scripts','plasma_store'
__requires__ = 'pyarrow==0.15.1'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('pyarrow==0.15.1', 'console_scripts', 'plasma_store')()
    )
